unit sablotengine;
{
/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is the Sablopas : Pascal wrapper for Sablotron.
 * More infos at http://www.tekool.com/sablopas - Contact : cespern@free.fr
 *
 * The Initial Developer of the Original Code is Christophe ESPERN
 * Portions created by Christophe ESPERN are Copyright (C) 2001-2002 Christophe
 * ESPERN. All Rights Reserved.
 *
 * Contributor(s): None
}

interface

{$IFDEF FPC}
 {$MODE DELPHI}
{$ENDIF}

uses
  SysUtils,Classes,shandler,sablot,sabutils;

const

  XML_HEADER='<?xml version="1.0"?>';
  XSL_HEADER= '<?xml version="1.0"?>'+
              '<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"'
              +' xmlns:xlink="http://www.w3.org/1999/xlink" >'
              +'</xsl:stylesheet>';

 XSL_FILE=0;
 XML_FILE=1;

type

  TSablotEngine=class;

  {Error Event}
  TSErrorEvent = procedure (Sender:TObject;const ErrorCode:integer;var 
          Fields:TSErrorFields) of object;


 {Scheme Events}
  TSFreeMemoryEvent = procedure (Sender:TObject;var buffer:PAnsiChar) of object;
  TSGetAllEvent = procedure (Sender:TObject;const scheme,rest:PAnsiChar;var
          buffer:PAnsiChar; var bytecount:integer) of object;
  TSCloseEvent = procedure (Sender:TObject;const handle:integer) of object;
  TSGetEvent = procedure (Sender:TObject;const handle:integer;var buffer:PAnsiChar;
          var bytecount:integer) of object;
  TSPutEvent = procedure (Sender:TObject;const handle:integer;const 
          buffer:PAnsiChar;var byteCount:integer) of object;
  TSOpenEvent = procedure (Sender:TObject;const scheme,rest:PAnsiChar;var
          handle:integer) of object;
  TSHasHandleEvent = procedure (Sender:TObject;handle:integer;var 
          Result:boolean) of object;
  {CustomSchemeProxy}

  TCustomSchemeProxy = class (TComponent)
  private
    FEngine: TSablotEngine;
    FSchemes: TStringlist;
    procedure SetEngine(Value: TSablotEngine);
  protected
    function Close(handle:integer): Integer; virtual; abstract;
    function FreeMemory(var buffer:PAnsiChar): Integer; virtual; abstract;
    function Get(const handle:integer;var buffer:PAnsiChar;var bytecount:integer):
            Integer; virtual; abstract;
    function GetAll(const scheme,rest:PAnsiChar;var buffer:PAnsiChar;var
            bytecount:integer): Integer; virtual; abstract;
    function HasHandle(handle:integer): Boolean; virtual; abstract;
    procedure Notification(AComponent: TComponent; Operation: TOperation); 
            override;
    function Open(const scheme,rest:PAnsiChar;var handle:integer): Integer; virtual;
            abstract;
    function Put(handle:integer;const buffer:PAnsiChar;var byteCount:integer):
            Integer; virtual; abstract;
  public
    constructor Create(Aowner:TComponent); override;
    destructor Destroy; override;
    procedure AddScheme(AScheme:Ansistring);
    function HasScheme(AScheme:Ansistring): Boolean;
    procedure RemoveScheme(AScheme:Ansistring);
  published
    property Engine: TSablotEngine read FEngine write SetEngine;
  end;

  {FileProxy}

  TFileProxy = class (TCustomSchemeProxy)
  private
    FHandles: TList;
    FInBuff: Ansistring;
    FChanged:boolean;
    FOpenMode:word;
  protected
    function Close(handle:integer): Integer; override;
    function FreeMemory(var buffer:PAnsiChar): Integer; override;
    function Get(const handle:integer;var buffer:PAnsiChar;var bytecount:integer):
            Integer; override;
    function GetAll(const scheme,rest:PAnsiChar;var buffer:PAnsiChar;var
            bytecount:integer): Integer; override;
    function HasHandle(handle:integer): Boolean; override;
    function Open(const scheme,rest:PAnsiChar;var handle:integer): Integer;
            override;
    function Put(handle:integer;const buffer:PAnsiChar;var byteCount:integer):
            Integer; override;
  public
    constructor Create(Aowner:TComponent); override;
    destructor Destroy; override;
  end;
  

  {SchemeProxy}
  TSchemeProxy = class (TCustomSchemeProxy)
  private
    FOnClose: TSCloseEvent;
    FOnFreeMemory: TSFreeMemoryEvent;
    FOnGet: TSGetEvent;
    FOnGetAll: TSGetAllEvent;
    FOnHasHandle: TSHasHandleEvent;
    FOnOpen: TSOpenEvent;
    FOnPut: TSPutEvent;
  protected
    function Close(handle:integer): Integer; override;
    function FreeMemory(var buffer:PAnsiChar): Integer; override;
    function Get(const handle:integer;var buffer:PAnsiChar;var bytecount:integer):
            Integer; override;
    function GetAll(const scheme,rest:PAnsiChar;var buffer:PAnsiChar;var
            bytecount:integer): Integer; override;
    function HasHandle(handle:integer): Boolean; override;
    function Open(const scheme,rest:PAnsiChar;var handle:integer): Integer;
            override;
    function Put(handle:integer;const buffer:PAnsiChar;var byteCount:integer):
            Integer; override;
  published
    property OnClose: TSCloseEvent read FOnClose write FOnClose;
    property OnFreeMemory: TSFreeMemoryEvent read FOnFreeMemory write 
            FOnFreeMemory;
    property OnGet: TSGetEvent read FOnGet write FOnGet;
    property OnGetAll: TSGetAllEvent read FOnGetAll write FOnGetAll;
    property OnHasHandle: TSHasHandleEvent read FOnHasHandle write FOnHasHandle;
    property OnOpen: TSOpenEvent read FOnOpen write FOnOpen;
    property OnPut: TSPutEvent read FOnPut write FOnPut;
  end;
  
  {SablotEngine}
  TSablotEngine = class (TComponent)
  private
    FBases: TStringlist;
    FDefaultBase: Ansistring;
    FFileProxy: TFileProxy;
    FInputUri: Ansistring;
    FLast   ug: Ansistring;
    FLastError: Ansistring;
    FMsgHandler: SMessageHandler;
    FOnMsgCritical: TSErrorEvent;
    FOnMsg   ug: TSErrorEvent;
    FOnMsgError: TSErrorEvent;
    FOnMsgInfo: TSErrorEvent;
    FOnMsgWarning: TSErrorEvent;
    FOutputUri: Ansistring;
    FParams: TStringList;
    FProcessor: PProcessor;
    FProxies: TList;
    FSchemeHandler: SSchemeHandler;
    FShow   ug: Boolean;
    FSituation: PSituation;
    FStyleURI: Ansistring;
    FTempProxy: TCustomSchemeProxy;
    function GetBaseByScheme(Scheme:Ansistring): Ansistring;
    function GetBaseCount: Integer;
    function GetProxies(index:integer): TCustomSchemeProxy;
    function GetProxyByHandle(AHandle:integer): TCustomSchemeProxy;
    function GetProxyCount: Integer;
    procedure InitSchemeHandler;

  protected
      procedure InternalClearBases;
    procedure PrepareBases;
    procedure PrepareParams;
    procedure AddProxy(AProxy:TCustomSchemeProxy);
    procedure ClearBases;
    procedure ClearError;
    procedure ClearParams;
    procedure DoRaiseEvent(code:MH_ERROR;level:MH_LEVEL;var 
            fields:TSErrorFields); virtual;
    function GetDefaultBase: Ansistring;
    procedure RemoveProxy(AProxy:TCustomSchemeProxy);
    function SabCreatePDocument(d:PDomDocument): Integer;
    function SabDestroyPDocument(d:SDomDocument): Integer;
    function SabLockDocument(d:SDomDocument): Integer;
    function SabParse(URI:Ansistring;d:PDomDocument): Integer;
    function SabParseBuffer(const buffer:Ansistring;d:PDomDocument): Integer;
    function SabParseStylesheet(URI:Ansistring;d:PDomDocument): Integer;
    function SabParseStylesheetBuffer(const buffer:Ansistring;d:PDomDocument):
            Integer;
    procedure SetDefaultBase(const ABase: Ansistring);
    property FileProxy: TFileProxy read FFileProxy;
    property Proxies[index:integer]: TCustomSchemeProxy read GetProxies;
    property ProxyCount: Integer read GetProxyCount;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    function AddArgBuffer(const argName,bufferValue:Ansistring): Boolean;
    procedure FillBaseList(strings:TStrings);
    procedure FillParamList(Strings:TStrings);
    procedure GetBasesList(strings:TStrings);
    procedure GetParamList(Strings:TStrings);
    function GetResultArg(argURI:Ansistring): Ansistring;
    function HasParam(paramName:Ansistring): Boolean;
    procedure RaiseMessage(level:MH_LEVEL;ModuleName,Msg: Ansistring); virtual;
    procedure RemoveBase(const scheme:Ansistring); virtual;
    procedure RemoveParam(const paramName:Ansistring); virtual;
    procedure ResolveSabURI(var Scheme,Rest:Ansistring);
    function Run: Boolean; virtual;
    function RunProcessorGen(AStyleURI,AInputURI,AOutputURI:Ansistring): Boolean;
    procedure SetBase(const scheme,base:Ansistring);
    procedure SetParam(const paramName,paramValue:Ansistring);
    property BaseByScheme[Scheme:Ansistring]: Ansistring read GetBaseByScheme;
    property BaseCount: Integer read GetBaseCount;
    property Last   ug: Ansistring read FLast   ug;
    property LastError: Ansistring read FLastError;
    property Processor: PProcessor read FProcessor;
    property Situation: PSituation read FSituation;
  published
    property DefaultBase: Ansistring read GetDefaultBase write SetDefaultBase;
    property InputUri: Ansistring read FInputUri write FInputUri;
    property OnMsgCritical: TSErrorEvent read FOnMsgCritical write 
            FOnMsgCritical;
    property OnMsg   ug: TSErrorEvent read FOnMsg   ug write FOnMsg   ug;
    property OnMsgError: TSErrorEvent read FOnMsgError write FOnMsgError;
    property OnMsgInfo: TSErrorEvent read FOnMsgInfo write FOnMsgInfo;
    property OnMsgWarning: TSErrorEvent read FOnMsgWarning write FOnMsgWarning;
    property OutputUri: Ansistring read FOutputUri write FOutputUri;
    property Show   ug: Boolean read FShow   ug write FShow   ug;
    property StyleURI: Ansistring read FStyleURI write FStyleURI;
  end;
  
procedure Register;


implementation

{$I handlers.inc}


procedure Register;
begin
  RegisterComponents('Tekool', [TSablotEngine]);
  RegisterComponents('Tekool', [TSchemeProxy]);
end;

{ TSchemeProxy }
{
****************************** TCustomSchemeProxy ******************************
}
constructor TCustomSchemeProxy.Create(Aowner:TComponent);
begin
  inherited;
  FSchemes:=TStringList.create;
  FSchemes.Duplicates:=dupError;
end;

destructor TCustomSchemeProxy.Destroy;
begin
  SetEngine(nil);
  FSchemes.free;
  inherited;
end;

procedure TCustomSchemeProxy.AddScheme(AScheme:Ansistring);
begin
  if FSchemes.IndexOf(AScheme)=-1 then
     FSchemes.Add(AScheme)
end;

function TCustomSchemeProxy.HasScheme(AScheme:Ansistring): Boolean;
begin
  if FSchemes.IndexOf(AScheme)<>-1 then
    Result:=true
  else
    Result:=false;
end;

procedure TCustomSchemeProxy.Notification(AComponent: TComponent; Operation: 
        TOperation);
begin
  if (AComponent=FEngine) and (Operation=opRemove) then
   SetEngine(nil);
  inherited;
end;

procedure TCustomSchemeProxy.RemoveScheme(AScheme:Ansistring);
begin
  if FSchemes.IndexOf(AScheme)<>-1 then
   FSchemes.Delete(FSchemes.IndexOf(AScheme));
end;

procedure TCustomSchemeProxy.SetEngine(Value: TSablotEngine);
begin
  if FEngine<>value then begin
   if FEngine<>nil then
     FEngine.RemoveProxy(self);
    FEngine := Value;
   if FEngine<>nil then
     FEngine.AddProxy(self)
  end
end;

{
********************************* TSchemeProxy *********************************
}
function TSchemeProxy.Close(handle:integer): Integer;
begin
  if Assigned(FOnClose) then
   FOnClose(self,handle);
   Result:=0;
end;

function TSchemeProxy.FreeMemory(var buffer:PAnsiChar): Integer;
begin
  Result:=1;
  if Assigned(FOnFreeMemory) then
   FOnFreeMemory(self,buffer);
  if buffer=nil then
   Result:=0
end;

function TSchemeProxy.Get(const handle:integer;var buffer:PAnsiChar;var
        bytecount:integer): Integer;
begin
  Result:=1;
  if Assigned(FOnGet) then
   FOnGet(self,handle,buffer,bytecount);
  if bytecount<>0 then
   Result:=0;
end;

function TSchemeProxy.GetAll(const scheme,rest:PAnsiChar;var buffer:PAnsiChar;var
        bytecount:integer): Integer;
var
  newRest:AnsiString;
begin
  //SetString(newRest, Rest, length(Rest));
  if Assigned(FOnGetAll) then
    FOnGetAll(self,scheme,rest,buffer,bytecount);
  if bytecount=0 then
   Result:=1
  else
   Result:=0;
end;

function TSchemeProxy.HasHandle(handle:integer): Boolean;
begin
  Result:=false;
  if Assigned(FOnHasHandle) then
   FOnHasHandle(self,handle,result);
end;

function TSchemeProxy.Open(const scheme,rest:PAnsiChar;var handle:integer): Integer;
begin
  Result:=1;
  if Assigned(FOnOpen) then
    FOnOpen(self,scheme,rest,handle);
  if handle<>0 then
   result:=0;
end;

function TSchemeProxy.Put(handle:integer;const buffer:PAnsiChar;var
        byteCount:integer): Integer;
begin
  Result:=1;
  if Assigned(FOnPut) then begin
   FOnPut(self,handle,buffer,bytecount);
   if bytecount=0 then
    Result:=1
   else
    Result:=0;
  end
end;

{
******************************** TSablotEngine *********************************
}
constructor TSablotEngine.Create(AOwner:TComponent);
begin
  inherited;
  FProxies:=TList.create;
  FFileProxy:=TFileProxy.Create(self);
  FFileProxy.engine:=self;
  FParams:=TStringlist.create;
  FBases:=TStringlist.create;
  FShow   ug:=false;
  //Create SablotSituation and SablotProcessor
  GetMem(FProcessor,SizeOf(PSablotHandle));
  GetMem(FSituation,SizeOf(PSablotHandle));
  SablotCreateSituation(FSituation);
  SablotCreateProcessorForSituation(FSituation^,FProcessor);
  //Init and record the Message Handler
  FMsgHandler.makeCode:=@SablotEngine.MessageHandlerMakeCode;
  FMsgHandler.log:=@SablotEngine.MessageHandlerLog;
  FMsgHandler.error:=@SablotEngine.MessageHandlerError;
  SablotRegHandler(FProcessor^,HLR_MESSAGE,@FMsgHandler,self);
  InitSchemeHandler;
end;

destructor TSablotEngine.Destroy;
begin
  SablotUnRegHandler(FProcessor^,HLR_MESSAGE,@FMsgHandler,self);
  SablotUnRegHandler(FProcessor^,HLR_SCHEME,@FSchemeHandler,self);
  SablotDestroyProcessor(FProcessor^);
  SablotDestroySituation(FSituation^);
  FreeMem(FProcessor,SizeOf(PSablotHandle));
  FreeMem(FSituation,SizeOf(PSablotHandle));
  inherited;
  FParams.free;
  FBases.free;
  FProxies.free;
end;

function TSablotEngine.AddArgBuffer(const argName,bufferValue:Ansistring): Boolean;
begin
  try
   if SablotAddArgBuffer(FSituation^,FProcessor^,PAnsiChar(ArgName),PAnsiCHar(BufferValue))=0 then
    Result:=true
   else
    Result:=false;
  except
   raise Exception.Create('TSablotEngine.AddArgBuffer');
  end;
end;

procedure TSablotEngine.AddProxy(AProxy:TCustomSchemeProxy);
begin
  FProxies.Add(AProxy);
end;

procedure TSablotEngine.ClearBases;
begin
  InternalClearBases;
  FBases.clear;
end;

procedure TSablotEngine.ClearError;
begin
  try
   SablotClearError(Fprocessor^);
   SablotClearSituation(FSituation^);
  except
   raise Exception.Create('TSablotEngine.Clear Error');
  end;
end;

procedure TSablotEngine.ClearParams;
begin
  FParams.Clear;
end;

procedure TSablotEngine.DoRaiseEvent(code:MH_ERROR;level:MH_LEVEL;var
        fields:TSErrorFields);
begin
  case level of
   MH_LEVEL_DEBUG:if Assigned(FOnMsg   ug)then
     FOnMsg   ug(self,code,fields);
   MH_LEVEL_INFO:if Assigned(FOnMsgInfo)then
     FOnMsgInfo(self,code,fields);
   MH_LEVEL_WARN:if Assigned(FOnMsgWarning)then
     FOnMsgWarning(self,code,fields);
   MH_LEVEL_ERROR:if Assigned(FOnMsgError)then
     FOnMsgError(self,code,fields);
   MH_LEVEL_CRITICAL:if Assigned(FOnMsgCritical)then
     FOnMsgCritical(self,code,fields);
  end;
end;

procedure TSablotEngine.FillBaseList(strings:TStrings);
begin
  FBases.Assign(strings);
end;

procedure TSablotEngine.FillParamList(Strings:TStrings);
begin
  FParams.Assign(strings);
end;

function TSablotEngine.GetBaseByScheme(Scheme:Ansistring): Ansistring;
var
  i: Integer;
  s1, s2: Ansistring;
begin
  for i:=0 to FBases.count-1 do begin
   SplitKeyValue(FBases[i],s1,s2);
   if s1=scheme then begin
     Result:=s2;
     exit;
   end;
  end;
end;

function TSablotEngine.GetBaseCount: Integer;
begin
  Result:=FBases.Count;
end;

procedure TSablotEngine.GetBasesList(strings:TStrings);
begin
  Strings.Assign(FBases);
end;

function TSablotEngine.GetDefaultBase: Ansistring;
begin
  Result:=FDefaultbase;
end;

procedure TSablotEngine.GetParamList(Strings:TStrings);
begin
  Strings.Assign(FParams);
end;

function TSablotEngine.GetProxies(index:integer): TCustomSchemeProxy;
begin
  if (index>-1) and (index<FProxies.count) then
   Result:= FProxies[index]
  else
   Result:=nil;
end;

function TSablotEngine.GetProxyByHandle(AHandle:integer): TCustomSchemeProxy;
var
  i: Integer;
begin
  Result:=nil;
  if  FileProxy.HasHandle(Ahandle) then
   Result:=FileProxy
  else
  with FProxies do
   for i:=0 to count-1 do
    if Proxies[i].HAsHandle(AHandle) then begin
     Result:=Proxies[i];
     break;
    end;
end;

function TSablotEngine.GetProxyCount: Integer;
begin
  Result:=FProxies.Count;
end;

function TSablotEngine.GetResultArg(argURI:Ansistring): Ansistring;
var
  tmp: PAnsiChar;
begin
  tmp:='';
  try
   SablotGetResultArg(Fprocessor^,PAnsichar(argUri),@tmp);
   Result:=Ansistring(tmp);
   SablotFree(tmp);
  except
   RaiseMessage(MH_LEVEL_DEBUG,'Sablopas','try to get an unknown output ?');
  end;
end;

function TSablotEngine.HasParam(paramName:Ansistring): Boolean;
var
  i: Integer;
  s1, s2: Ansistring;
begin
  Result:=false;
  for i:=0 to FParams.count-1 do begin
   SplitKeyValue(FParams[i],s1,s2);
   if s1=paramName then begin
    Result:=true;
     break;
   end;
  end;
end;

procedure TSablotEngine.InitSchemeHandler;
begin
  try
   FSchemeHandler.put:=@SablotEngine.SabSchemeHandlerPut;
   FSchemeHandler.get:=@SablotEngine.SabSchemeHandlerGet;
   FSchemeHandler.open:=@SablotEngine.SabSchemeHandlerOpen;
   FSchemeHandler.close:=@SablotEngine.SabSchemeHandlerClose;
   FSchemeHandler.getAll:=@SablotEngine.SabSchemeHandlerGetAll;
   FSchemeHandler.freeMemory:=@SablotEngine.SabSchemeHandlerFreeMemory;
   SablotRegHandler(FProcessor^,HLR_Scheme,@FSchemeHandler,self);
  except
   raise Exception.Create('TSablotEngine.InitSchemeHandler');
  end;
end;

procedure TSablotEngine.InternalClearBases;
var
  i: Integer;
  s1, s2: Ansistring;
begin
  try
   for i:=0 to FBases.count-1 do begin
    splitKeyValue(FBases[i],s1,s2);
    s2:='';
    if SablotSetBaseForScheme(FProcessor^,PAnsiCHar(s1),PAnsiCHar(s2))<>0 then
     RaiseMessage(MH_LEVEL_ERROR,'Sablopas','Cannot clear base for '+S1+' : '+S2);
   end;
  except
   raise Exception.Create('TSablotEngine.SetBaseForScheme error');
  end;
end;

procedure TSablotEngine.PrepareBases;
var
  i: Integer;
  s1, s2: Ansistring;
begin
  try
   for i:=0 to FBases.count-1 do begin
    splitKeyValue(FBases[i],s1,s2);
    if SablotSetBaseForScheme(FProcessor^,PAnsiCHar(s1),PAnsiCHar(s2))<>0 then
     RaiseMessage(MH_LEVEL_ERROR,'Sablopas','Cannot set scheme base for '+S1+' : '+S2);
   end;
  except
   raise Exception.Create('TSablotEngine.PrepareBases error');
  end;
end;

procedure TSablotEngine.PrepareParams;
var
  i: Integer;
  s1, s2: Ansistring;
begin
  for i:=0 to FParams.count-1 do begin
   splitKeyValue(FParams[i],s1,s2);
   try
   if SablotAddParam(FSituation^,FProcessor^,PAnsiChar(s1),PAnsiCHar(s2))<>0 then
    RaiseMessage(MH_LEVEL_ERROR,'Sablopas','Cannot add param '+s1);
   except
    Exception.create('TSablotEngine.PrepareParams with '+s1+'='+s2);
  end;
  end;
end;

procedure TSablotEngine.RaiseMessage(level:MH_LEVEL;ModuleName,Msg: Ansistring);
var
  Fields: TSErrorFields;
begin
  
  Fields[0]:=PAnsiChar('module:'+ModuleName);
  Fields[2]:=PAnsiChar('code: 100' );
  Fields[3]:=PAnsiChar('msg: '+Msg);
  case level of
   MH_LEVEL_DEBUG: Fields[1]:='msgtype: debug';
   MH_LEVEL_INFO: Fields[1]:='msgtype: info';
   MH_LEVEL_WARN:Fields[1]:='msgtype: warning';
   MH_LEVEL_ERROR:Fields[1]:='msgtype: error';
   MH_LEVEL_CRITICAL:Fields[1]:='msgtype: critical';
  end;
  Fields[4]:=nil;
  MessageHandlerError(self,self.FProcessor^,100,level,fields);
end;

procedure TSablotEngine.RemoveBase(const scheme:Ansistring);
var
  i: Integer;
  s1, s2: Ansistring;
begin
  for i:=0 to FBases.count-1 do begin
   SplitKeyValue(FBases[i],s1,s2);
   if s1=scheme then begin
     FBases.delete(i);
     if SablotSetBaseForScheme(FProcessor^,PAnsiCHar(s1),'')<>0 then
       Exception.create('Cannot remove base for '+S1+' : '+S2);
     break;
   end;
  end;
end;

procedure TSablotEngine.RemoveParam(const paramName:Ansistring);
var
  i: Integer;
  s1, s2: Ansistring;
begin
  for i:=0 to FParams.count-1 do begin
   SplitKeyValue(FParams[i],s1,s2);
   if s1=paramName then begin
     Fparams.delete(i);
     break;
   end;
  end;
end;

procedure TSablotEngine.RemoveProxy(AProxy:TCustomSchemeProxy);
begin
  if  FProxies.IndexOf(AProxy)<>-1 then
   FProxies.Remove(AProxy)
end;

procedure TSablotEngine.ResolveSabURI(var Scheme,Rest:Ansistring);
var
  i: Integer;
  s1, s2, authority, query, fragment, uri: Ansistring;
  filename: tfilename;
begin
  for i:=0 to FBases.count-1 do begin
   SplitKeyValue(FBases[i],s1,s2);
   if s1=scheme then begin
       //Change the scheme
     s1:=Copy(s2,0,pos('://',s2)-1);
     //Get the filename
     if s1='file' then
       UriStrToFilename(s1+':'+rest,filename,authority,query,fragment) // Conversion localfile
     else
      filename:=copy(rest,pos('//',rest)+2,length(rest)); //Enlve le double /
     filename:=DosPAthToUnixPath(filename);
     //URI Creation
     ResolveRelativeURIStr(s2,authority+'/'+filename,uri);
     if Copy(uri,length(uri),1)='/' then
      uri:=Copy(uri,0,length(uri)-1);
     //URI Split
     Rest:=Copy(uri,pos('://',uri),length(uri));
     Scheme:=s1;
     break;
   end;
  end;
end;

function TSablotEngine.Run: Boolean;
begin
  Result:= RunProcessorGen(FstyleUri,FInputUri,FOutputUri);
end;

function TSablotEngine.RunProcessorGen(AStyleURI,AInputURI,AOutputURI:Ansistring):
        Boolean;
begin
  try
    PrepareBases;
    PrepareParams;
  
    if (AStyleURI='') or (AInputURI='') or (AOutputURI='') then
     raise Exception.Create('RunProcessorGen : Missing file argument');
    if SablotRunProcessorGen(FSituation^,FProcessor^,PAnsichar(AStyleUri),PAnsiCHar(AInputUri),PAnsiChar(AOutputUri))=0 then
      Result:=true
    else begin
      ClearError;
      SablotFreeResultArgs(FProcessor^);
      Result:=false;
    end;
  
    InternalClearBases;
  except
   on E:Exception do begin
    RaiseMessage(MH_LEVEL_ERROR,'Sablopas',(E.Message));
    Result:=false;
    InternalClearBases;
   end;
  end;
  
end;

function TSablotEngine.SabCreatePDocument(d:PDomDocument): Integer;
begin
  try
   Result:=SablotCreateDocument(FSituation^,d);
  except
    raise Exception.Create('TSablotEngine.DestroyPDocument');
  end;
end;

function TSablotEngine.SabDestroyPDocument(d:SDomDocument): Integer;
begin
  try
   Result:=SablotDestroyDocument(FSituation^,d)
  except
    raise Exception.Create('TSablotEngine.DestroyPDocument');
  end;
end;

function TSablotEngine.SabLockDocument(d:SDomDocument): Integer;
begin
  Result:=SablotLockDocument(FSituation^,d);
end;

function TSablotEngine.SabParse(URI:Ansistring;d:PDomDocument): Integer;
begin
  Result:=SablotParse(FSituation^,PAnsiCHar(URI),d);
end;

function TSablotEngine.SabParseBuffer(const buffer:Ansistring;d:PDomDocument):
        Integer;
begin
  Result:=SablotParseBuffer(FSituation^,PAnsiCHar(buffer),d);
end;

function TSablotEngine.SabParseStylesheet(URI:Ansistring;d:PDomDocument): Integer;
begin
  Result:=SablotParseStyleSheet(FSituation^,PAnsiChar(URI),d);
end;

function TSablotEngine.SabParseStylesheetBuffer(const buffer:Ansistring;
        d:PDomDocument): Integer;
begin
  Result:=SablotParseStyleSheetBuffer(FSituation^,PAnsiChar(buffer),d);
end;

procedure TSablotEngine.SetBase(const scheme,base:Ansistring);
var
  i: Integer;
  s1, s2: Ansistring;
begin
  for i:=0 to FBases.count-1 do begin
   SplitKeyValue(FBases[i],s1,s2);
   if s1=scheme then begin
     FBases[i]:=scheme+'='+base;
     exit;
   end;
  end;
  if scheme<>'' then
    FBases.Add(scheme+'='+base)
end;

procedure TSablotEngine.SetDefaultBase(const ABase: Ansistring);
begin
  FDefaultBase:=ABase;
  SablotSetBase(FProcessor^,PAnsiChar(FDefaultBase));
end;

procedure TSablotEngine.SetParam(const paramName,paramValue:Ansistring);
var
  i: Integer;
  s1, s2: Ansistring;
begin
  for i:=0 to FParams.count-1 do begin
   SplitKeyValue(FParams[i],s1,s2);
   if s1=paramName then begin
     FParams[i]:=paramName+'='+paramvalue;
     exit;
   end;
  end;
  if paramName<>'' then
    FParams.Add(paramName+'='+paramvalue)
end;

{ TFileProxy }

{
********************************** TFileProxy **********************************
}
constructor TFileProxy.Create(Aowner:TComponent);
begin
  FHandles:=TList.Create;
  inherited;
end;

destructor TFileProxy.Destroy;
begin
  FHandles.Free;
  inherited;
end;

function TFileProxy.Close(handle:integer): Integer;
begin
  try
   FHandles.Remove(pointer(handle));
   TObject(handle).free;
   Result:=0;
  except
   Result:=1;
  end;
end;

function TFileProxy.FreeMemory(var buffer:PAnsiChar): Integer;
begin
  FInBuff:='';
  Result:=0;
end;

function TFileProxy.Get(const handle:integer;var buffer:PAnsiChar;var
        bytecount:integer): Integer;
begin
  with TStream(handle) do
  try
   ReadBuffer(buffer,size);
   bytecount:=size;
   Result:=0;
  except
   Result:=0;
  end;
end;

function TFileProxy.GetAll(const scheme,rest:PAnsiChar;var buffer:PAnsiChar;var
        bytecount:integer): Integer;
var
  fname: Ansistring;
  FStream: TFileStream;
begin
  Result:=1;
  try
   fname:=Copy(rest,Pos('://',rest)+3,length(rest));
   if not FileExists(fname) then
    exit;
   FStream:=TFileStream.Create(fname,fmOpenRead);
   SetLength(FInBuff,fstream.size);
   FStream.ReadBuffer(pointer(FInBuff)^,Fstream.size);
   buffer:=PAnsiChar(FInbuff);
   bytecount:=strlen(buffer);
   fstream.free;
   Result:=0;
  except
   if Assigned(FStream) then
    FStream.Free;
   FInBuff:='';
  end
end;

function TFileProxy.HasHandle(handle:integer): Boolean;
begin
  if  FHandles.IndexOf(Pointer(handle))<>-1 then
   Result:=true
  else
   Result:=false;
end;

function TFileProxy.Open(const scheme,rest:PAnsiChar;var handle:integer): Integer;
var
  fname: Ansistring;
begin
  Result:=1;
  fname:=Copy(rest,Pos('://',rest)+3,length(rest));
  if FileExists(fname) then
   FOpenMode:=fmOpenwrite
  else
   FOpenMode:=fmCreate;
  FHandles.Add(TFileStream.Create(fname,FOpenMode));
  handle:=integer(FHandles[Fhandles.Count-1]);
  if handle<>0 then
   Result:=0;
end;

function TFileProxy.Put(handle:integer;const buffer:PAnsiChar;var
        byteCount:integer): Integer;
begin
  try
  with TStream(handle) do begin
   WriteBuffer(buffer^,bytecount);
   bytecount:=bytecount+1;
   Result:=0;
  end;
  except
   Result:=1;
  end;
end;

end.




