unit uROR_NamespaceHandler;
{
================================================================================
*
*       Package:        ROR - Clinical Case Registries
*       Date Created:   $Revision: 9 $  $Modtime: 3/07/07 3:47p $
*       Site:           Hines OIFO
*       Developers:
*                                      
*
*       Description:    Namespace Handler for the embedded web browser
*
*       Notes:
*
================================================================================
*       $Archive: /CCR v1.5/Current/uROR_NamespaceHandler.pas $
*
* $History: uROR_NamespaceHandler.pas $
 * 
 * *****************  Version 9  *****************
 * User: Vhaishgavris Date: 3/08/07    Time: 1:47p
 * Updated in $/CCR v1.5/Current
 * 
 * *****************  Version 8  *****************
 * User: Vhaishgavris Date: 10/13/06   Time: 11:35a
 * Updated in $/CCR v1.5/Current
 * 
 * *****************  Version 7  *****************
 * User: Vhaishgavris Date: 1/10/05    Time: 3:48p
 * Updated in $/CCR v1.0/Current
 * 
 * *****************  Version 6  *****************
 * User: Vhaishgavris Date: 10/14/04   Time: 3:50p
 * Updated in $/CCR v1.0/Current
 * 
 * *****************  Version 5  *****************
 * User: Vhaishgavris Date: 6/25/04    Time: 3:55p
 * Updated in $/CCR v1.0/Current
 * 
 * *****************  Version 4  *****************
 * User: Vhaishgavris Date: 9/22/03    Time: 7:48a
 * Updated in $/ICR v3.0/Current
 * 
 * *****************  Version 3  *****************
 * User: Vhaishgavris Date: 9/19/03    Time: 3:54p
 * Updated in $/ICR v3.0/Current
 *
 * *****************  Version 2  *****************
 * User: Vhaishgavris Date: 9/11/03    Time: 10:55p
 * Updated in $/ICR v3.0/Current
 *
 * *****************  Version 1  *****************
 * User: Vhaishgavris Date: 9/11/03    Time: 5:03p
 * Created in $/ICR v3.0/Current
*
================================================================================
}
interface
uses
  Classes, WinApi.Windows, SysUtils, System.Win.ComObj, WinApi.UrlMon;

const
  Class_OurNSHandler: TGUID = '{A3DB4CB3-4683-4E39-8073-AE3430C7746C}';
  sReportNameSpace = 'http://vista/report/';
  sTemplateNameSpace = 'http://vista/template/';

type
  TOurNSHandler = class(TComObject, IInternetProtocol)

  private
    ProtSink: IInternetProtocolSink;
    dataStream: TStringStream;

    function loadTemplate( aCode, aType: String ): TStringStream;
    function transformReport( aTask: String; aSortMode: TStringList ): TStringStream;

  protected
    // IInternetProtocol Methods
    function Start(szUrl: PWideChar; OIProtSink: IInternetProtocolSink;
      OIBindInfo: IInternetBindInfo; grfPI, dwReserved: DWORD): HResult; stdcall;
    function Continue(const ProtocolData: TProtocolData): HResult; stdcall;
    function Abort(hrReason: HResult; dwOptions: DWORD): HResult; stdcall;
    function Terminate(dwOptions: DWORD): HResult; stdcall;
    function Suspend: HResult; stdcall;
    function Resume: HResult; stdcall;
    function Read(pv: Pointer; cb: ULONG; out cbRead: ULONG): HResult; stdcall;
    function Seek(dlibMove: LARGE_INTEGER; dwOrigin: DWORD;
      out libNewPosition: ULARGE_INTEGER): HResult; stdcall;
    function LockRequest(dwOptions: DWORD): HResult; stdcall;
    function UnlockRequest: HResult; stdcall;

  public

    class function  Register: Boolean;
    class procedure Unregister;

  end;

function ParseReportURL(URL: String; var Mode: String;
                         var Task: String; var SortMode: TStringList): Boolean;

procedure SortModeFromURL(SortMode:TStringList; SortParams: String);

implementation

uses
  comserv, uROR_Utilities, uROR_Common, fROR_XMLEngine, ActiveX, uROR_Broker;

var
  ClassFactory: IClassFactory;
  InternetSession: IInternetSession;


function ParseReportURL(URL: String; var Mode: String;
                         var Task: String; var SortMode: TStringList): Boolean;
var
  buf: String;
begin
  buf := LowerCase(URL);
  if Pos(sReportNameSpace, buf) = 1 then
    begin
      buf := Piece(Piece(buf,sReportNameSpace,2),'?');
      Mode := Piece(buf, '/', 1);
      Task := Piece(buf, '/', 2);
      if Not Assigned(SortMode) then
        SortMode := TStringList.Create;
      SortModeFromURL(SortMode, Piece(URL,'?',2,999));
      Result := True;
    end
  else
    begin
      Mode := ''; Task := '';
      Result := False;
    end;
end;

procedure SortModeFromURL(SortMode: TStringList; SortParams: String);
var
  sortParent, param, sortElement: String;
  i, n: Integer;
begin
  i := 1;
  param := Piece(SortParams, '&', i);
  while param <> '' do
    begin
      sortParent := Trim( Piece(param, '=') );
      sortElement := Trim( Piece(param, '=', 2) );
      n := SortMode.IndexOfName(sortParent);
      if n >= 0 then
        begin
          //--- Toggle the sort order
          if SortMode.Values[sortParent] = sortElement then
            if Pos('DESC', Piece(sortElement, ':', 2)) > 0 then
              sortElement := Piece(sortElement, ':')
            else
              sortElement := sortElement + ':DESC';
          //--- Remove the ASC modifier
          if Pos('ASC', Piece(sortElement, ':', 2)) > 0 then
            sortElement := Piece(sortElement, ':');
          //--- Update the sort mode
          SortMode.Strings[n] := sortParent + '=' + sortElement;
        end
      else
        SortMode.Add(sortParent + '=' + sortElement);
      Inc(i);
      param := Piece(SortParams, '&', i);
    end;
end;

function TOurNSHandler.Abort(hrReason: HResult; dwOptions: DWORD): HResult; stdcall;
begin
  Result := E_NOTIMPL;
end;

function TOurNSHandler.Continue(const ProtocolData: TProtocolData): HResult;
begin
  Result := S_OK;
end;

function TOurNSHandler.loadTemplate( aCode, aType: String ): TStringStream;
var
  resBuf: TStringList;
begin
  Result := TStringStream.Create('');
  resBuf := TStringList.Create;
  try
    if Broker.CallProc(rpcReportStylesheet, [aCode, aType], nil, resBuf) then
      resBuf.SaveToStream(Result);
  finally
    resBuf.Free;
  end;
end;

function TOurNSHandler.LockRequest(dwOptions: DWORD): HResult; stdcall;
begin
  Result := S_OK;
end;

function TOurNSHandler.Read(pv: Pointer; cb: ULONG; out cbRead: ULONG): HResult;
begin
  // Read Data from DataStream to Browser/URLMON
  cbRead := dataStream.Read(pv^, cb);
  if cbRead < cb then
    Result := S_FALSE
  else
    Result := S_OK;
end;

class function TOurNSHandler.Register: Boolean;
begin
  Result := False;
  try
    // Register the namespace for embedded web browser
    CoGetClassObject(Class_OurNSHandler, CLSCTX_SERVER, nil, IClassFactory, ClassFactory);
    ClassFactory._AddRef;
    CoInternetGetSession(0, InternetSession, 0);
    if Assigned(InternetSession) then
      begin
        InternetSession._AddRef;
        if InternetSession.RegisterNameSpace(ClassFactory,
                Class_OurNSHandler, 'http', 0, nil, 0) = S_OK then
          Result := True;
      end;
  except
  end;
end;

function TOurNSHandler.Resume: HResult; stdcall;
begin
  result := E_NOTIMPL;
end;

function TOurNSHandler.Seek(dlibMove: LARGE_INTEGER; dwOrigin: DWORD;
  out libNewPosition: ULARGE_INTEGER): HResult;
begin
  result := E_NOTIMPL;
end;

{------------------------------------------------------------
 URL: http://vista/report/task/12345?PATIENTS=NAME:DESC&...
      http://vista/report/sort?PATIENTS=NAME:DESC&...
      http://vista/template/5/1
 ------------------------------------------------------------}

function TOurNSHandler.Start(szUrl: PWideChar; OIProtSink: IInternetProtocolSink;
  OIBindInfo: IInternetBindInfo; grfPI, dwReserved: DWORD): HResult; stdcall;
var
  buf, lcUrl, mode, task: String;
  TotalSize: Longint;
  SortMode: TStringList;
begin
  SortMode := nil;
  if ParseReportURL(szUrl, mode, task, SortMode) then
    begin
      dataStream := transformReport(task, SortMode);
      SortMode.Free;
      Result := S_OK;
    end
  else
    begin
      lcUrl := AnsiLowerCase(szURL);
      if Pos(sTemplateNameSpace, lcUrl) = 1 then
        begin
          buf := Piece(lcUrl, sTemplateNameSpace, 2);
          dataStream := loadTemplate(Piece(buf,'/'), Piece(buf,'/',2));
          Result := S_OK;
        end
      else
        Result := INET_E_USE_DEFAULT_PROTOCOLHANDLER;
    end;

  if Result = S_OK then
    begin
      TotalSize := dataStream.position;
      dataStream.Seek(0, soFromBeginning);
      //Get interface to Transaction handlers IInternetnetProtocolSink
      ProtSink := OIProtSink;
      // Inform Transaction handler that all data is ready
      ProtSink.ReportData(BSCF_FIRSTDATANOTIFICATION or
            BSCF_LASTDATANOTIFICATION or BSCF_DATAFULLYAVAILABLE, TotalSize, TotalSize);
      // Here our Read Method is called by transaction handler
      ProtSink.ReportResult(S_OK, 0, nil);
    end;
end;

function TOurNSHandler.Suspend: HResult; stdcall;
begin
  result := E_NOTIMPL;
end;

function TOurNSHandler.Terminate(dwOptions: DWORD): HResult; stdcall;
begin
  dataStream.Free;
  ProtSink._Release;
  Result := S_OK;
end;

function TOurNSHandler.transformReport( aTask: String;
                aSortMode: TStringList ): TStringStream;
var
  engine: TXMLEngine;
  empty: Boolean;
begin
  Result := TStringStream.Create('');
  engine := TXMLEngine.Create(Broker);

  if Not engine.TransformReport(aTask, Result, '', aSortMode) then
    with Result do
      begin
        Seek(0, soFromEnd);
        empty := (Position = 0);
        if empty then WriteString(sOpenHTML);
        WriteString('<div><pre>');
        engine.Errors.SaveToStream(Result);
        WriteString('</pre></div>');
        if empty then WriteString(sCloseHTML);
      end;

  engine.Free;
end;

function TOurNSHandler.UnlockRequest: HResult;
begin
  result := S_OK;
end;

class procedure TOurNSHandler.Unregister;
begin
  if Assigned(InternetSession) then
    begin
      InternetSession.UnregisterNameSpace(ClassFactory, 'http');
      InternetSession._Release;
      InternetSession := nil;
    end;
  if Assigned(ClassFactory) then
    begin
      ClassFactory._Release;
      ClassFactory := nil;
    end;
end;

initialization
  TComObjectFactory.Create(ComServer, TOurNSHandler, Class_OurNSHandler,
    'OurNSHandler', 'OurNSHandler', ciMultiInstance, tmApartment);

end.
